{------------------------------------------------------------------------------}
{                                                                              }
{  TGradient v2.522                                                            }
{  by Kambiz R. Khojasteh                                                      }
{                                                                              }
{  kambiz@delphiarea.com                                                       }
{  http://www.delphiarea.com                                                   }
{                                                                              }
{------------------------------------------------------------------------------}

{$I DELPHIAREA.INC}

unit Gradient;

interface

uses
  Windows, Messages, Classes, Graphics, Controls, Forms, Dialogs, Menus;

type

  PRGBTripleArray = ^TRGBTripleArray;
  TRGBTripleArray = array[0..1024] of TRGBTriple;

  TGradientColors = array[0..255] of TRGBTriple;

  TGradientShift = -100..100;
  TGradientRotation = -100..100;

  TGradientStyle = (gsCustom, gsRadialC, gsRadialT, gsRadialB, gsRadialL,
    gsRadialR, gsRadialTL, gsRadialTR, gsRadialBL, gsRadialBR, gsLinearH,
    gsLinearV, gsReflectedH, gsReflectedV, gsDiagonalLF, gsDiagonalLB,
    gsDiagonalRF, gsDiagonalRB, gsArrowL, gsArrowR, gsArrowU, gsArrowD,
    gsDiamond, gsButterfly);

  TCustomGradientEvent = procedure(Sender: TObject; const Colors: TGradientColors;
    Pattern: TBitmap) of object;

  TGradient = class(TGraphicControl)
  private
    fColorBegin: TColor;
    fColorEnd: TColor;
    fUseSysColors: Boolean;
    fStyle: TGradientStyle;
    fShift: TGradientShift;
    fRotation: TGradientRotation;
    fReverse: Boolean;
    fPattern: TBitmap;
    fOnCustom: TCustomGradientEvent;
    UpdateCount: Integer;
    UpdatePended: Boolean;
    Dirty: Boolean;
    procedure SetColorBegin(Value: TColor);
    procedure SetColorEnd(Value: TColor);
    procedure SetUseSysColors(Value: Boolean);
    procedure SetStyle(Value: TGradientStyle);
    procedure SetShift(Value: TGradientShift);
    procedure SetRotation(Value: TGradientRotation);
    procedure SetReverse(Value: Boolean);
    function IsColorBeginSaved: Boolean;
    function IsColorEndSaved: Boolean;
    procedure WMSETTINGCHANGE(var Message: TMessage); message WM_SETTINGCHANGE;
  protected
    procedure Paint; override;
    procedure Loaded; override;
    procedure UpdatePattern; virtual;
    procedure UpdateSysColors; virtual;
    property Pattern: TBitmap read fPattern;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function CopyPatternTo(Bitmap: TBitmap): Boolean;
    procedure InvalidatePattern;
    procedure BeginUpdate;
    procedure EndUpdate;
  published
    property Align;
    {$IFDEF DELPHI4_UP}
    property Anchors;
    {$ENDIF}
    property ColorBegin: TColor read fColorBegin write SetColorBegin stored IsColorBeginSaved;
    property ColorEnd: TColor read fColorEnd write SetColorEnd stored IsColorEndSaved;
    {$IFDEF DELPHI4_UP}
    property Constraints;
    {$ENDIF}
    property DragCursor;
    {$IFDEF DELPHI4_UP}
    property DragKind;
    {$ENDIF}
    property DragMode;
    property Enabled;
    property Height default 100;
    property ParentShowHint;
    property PopupMenu;
    property Reverse: Boolean read fReverse write SetReverse default False;
    property Rotation: TGradientRotation read fRotation write SetRotation default 0;
    property Shift: TGradientShift read fShift write SetShift default 0;
    property ShowHint;
    property Style: TGradientStyle read fStyle write SetStyle default gsRadialC;
    property UseSysColors: Boolean read fUseSysColors write SetUseSysColors default False;
    property Visible;
    property Width default 100;
    property OnClick;
    property OnCustom: TCustomGradientEvent read fOnCustom write fOnCustom;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    {$IFDEF DELPHI4_UP}
    property OnEndDock;
    {$ENDIF}
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    {$IFDEF DELPHI4_UP}
    property OnStartDock;
    {$ENDIF}
    property OnStartDrag;
  end;

procedure Register;

implementation

procedure RadialCentral(const Colors: TGradientColors; Pattern: TBitmap);
var
  X, Y: Integer;
  Row: PRGBTripleArray;
  PreCalcX, PreCalcY: Integer;
  PreCalcXs: array[0..361] of Integer;
begin
  Pattern.Width := 362;
  Pattern.Height := 362;
  for X := 0 to 180 do
  begin
    PreCalcX := 180 - X;
    PreCalcXs[X] := PreCalcX * PreCalcX;
  end;
  for X := 181 to 361 do
  begin
    PreCalcX := X - 181;
    PreCalcXs[X] := PreCalcX * PreCalcX;
  end;
  for Y := 0 to 361 do
  begin
    PreCalcY := 180 - Y;
    PreCalcY := PreCalcY * PreCalcY;
    Row := PRGBTripleArray(Pattern.ScanLine[Y]);
    for X := 0 to 361 do
      Row[X] := Colors[Round(Sqrt(PreCalcXs[X] + PreCalcY))];
  end;
  { Not optimized code
  for Y := 0 to 361 do
  begin
    Row := PRGBTripleArray(Pattern.ScanLine[Y]);
    for X := 0 to 180 do
      Row[X] := Colors[Round(Sqrt(Sqr(180 - X) + Sqr(180 - Y)))];
    for X := 181 to 361 do
      Row[X] := Colors[Round(Sqrt(Sqr(X - 181) + Sqr(180 - Y)))];
  end;
  }
end;

procedure RadialTop(const Colors: TGradientColors; Pattern: TBitmap);
var
  X, Y: Integer;
  Row: PRGBTripleArray;
  PreCalcX, PreCalcY: Integer;
  PreCalcXs: array[0..361] of Integer;
begin
  Pattern.Width := 362;
  Pattern.Height := 181;
  for X := 0 to 180 do
  begin
    PreCalcX := 180 - X;
    PreCalcXs[X] := PreCalcX * PreCalcX;
  end;
  for X := 181 to 361 do
  begin
    PreCalcX := X - 181;
    PreCalcXs[X] := PreCalcX * PreCalcX;
  end;
  for Y := 0 to 180 do
  begin
    PreCalcY := Y * Y;
    Row := PRGBTripleArray(Pattern.ScanLine[Y]);
    for X := 0 to 361 do
      Row[X] := Colors[Round(Sqrt(PreCalcXs[X] + PreCalcY))];
  end;
  { Not optimized code
  for Y := 0 to 180 do
  begin
    Row := PRGBTripleArray(Pattern.ScanLine[Y]);
    for X := 0 to 180 do
      Row[X] := Colors[Round(Sqrt(Sqr(180 - X) + Sqr(Y)))];
    for X := 181 to 361 do
      Row[X] := Colors[Round(Sqrt(Sqr(X - 181) + Sqr(Y)))];
  end;
  }
end;

procedure RadialBottom(const Colors: TGradientColors; Pattern: TBitmap);
var
  X, Y: Integer;
  Row: PRGBTripleArray;
  PreCalcX, PreCalcY: Integer;
  PreCalcXs: array[0..361] of Integer;
begin
  Pattern.Width := 362;
  Pattern.Height := 181;
  for X := 0 to 180 do
  begin
    PreCalcX := 180 - X;
    PreCalcXs[X] := PreCalcX * PreCalcX;
  end;
  for X := 181 to 361 do
  begin
    PreCalcX := X - 181;
    PreCalcXs[X] := PreCalcX * PreCalcX;
  end;
  for Y := 0 to 180 do
  begin
    PreCalcY := 180 - Y;
    PreCalcY := PreCalcY * PreCalcY;
    Row := PRGBTripleArray(Pattern.ScanLine[Y]);
    for X := 0 to 361 do
      Row[X] := Colors[Round(Sqrt(PreCalcXs[X] + PreCalcY))];
  end;
  { Not optimized code
  for Y := 0 to 180 do
  begin
    Row := PRGBTripleArray(Pattern.ScanLine[Y]);
    for X := 0 to 180 do
      Row[X] := Colors[Round(Sqrt(Sqr(180 - X) + Sqr(180 - Y)))];
    for X := 181 to 361 do
      Row[X] := Colors[Round(Sqrt(Sqr(X - 181) + Sqr(180 - Y)))];
  end;
  }
end;

procedure RadialLeft(const Colors: TGradientColors; Pattern: TBitmap);
var
  X, Y: Integer;
  Row: PRGBTripleArray;
  PreCalcY: Integer;
  PreCalcYs: array[0..361] of Integer;
begin
  Pattern.Width := 181;
  Pattern.Height := 362;
  for Y := 0 to 180 do
  begin
    PreCalcY := 180 - Y;
    PreCalcYs[Y] := PreCalcY * PreCalcY;
  end;
  for Y := 181 to 361 do
  begin
    PreCalcY := Y - 181;
    PreCalcYs[Y] := PreCalcY * PreCalcY;
  end;
  for Y := 0 to 361 do
  begin
    Row := PRGBTripleArray(Pattern.ScanLine[Y]);
    for X := 0 to 180 do
      Row[X] := Colors[Round(Sqrt(X * X + PreCalcYs[Y]))];
  end;
  { Not optimized code
  for Y := 0 to 180 do
  begin
    Row := PRGBTripleArray(Pattern.ScanLine[Y]);
    for X := 0 to 180 do
      Row[X] := Colors[Round(Sqrt(Sqr(X) + Sqr(180 - Y)))];
  end;
  for Y := 181 to 361 do
  begin
    Row := PRGBTripleArray(Pattern.ScanLine[Y]);
    for X := 0 to 180 do
      Row[X] := Colors[Round(Sqrt(Sqr(X) + Sqr(Y - 181)))];
  end;
  }
end;

procedure RadialRight(const Colors: TGradientColors; Pattern: TBitmap);
var
  X, Y: Integer;
  Row: PRGBTripleArray;
  PreCalcX, PreCalcY: Integer;
  PreCalcXs: array[0..180] of Integer;
  PreCalcYs: array[0..361] of Integer;
begin
  Pattern.Width := 181;
  Pattern.Height := 362;
  for X := 0 to 180 do
  begin
    PreCalcX := 180 - X;
    PreCalcXs[X] := PreCalcX * PreCalcX;
  end;
  for Y := 0 to 180 do
  begin
    PreCalcY := 180 - Y;
    PreCalcYs[Y] := PreCalcY * PreCalcY;
  end;
  for Y := 181 to 361 do
  begin
    PreCalcY := Y - 181;
    PreCalcYs[Y] := PreCalcY * PreCalcY;
  end;
  for Y := 0 to 361 do
  begin
    Row := PRGBTripleArray(Pattern.ScanLine[Y]);
    for X := 0 to 180 do
      Row[X] := Colors[Round(Sqrt(PreCalcXs[X] + PreCalcYs[Y]))];
  end;
  { Not optimized code
  for Y := 0 to 180 do
  begin
    Row := PRGBTripleArray(Pattern.ScanLine[Y]);
    for X := 0 to 180 do
      Row[X] := Colors[Round(Sqrt(Sqr(180 - X) + Sqr(180 - Y)))];
  end;
  for Y := 181 to 361 do
  begin
    Row := PRGBTripleArray(Pattern.ScanLine[Y]);
    for X := 0 to 180 do
      Row[X] := Colors[Round(Sqrt(Sqr(180 - X) + Sqr(Y - 181)))];
  end;
  }
end;

procedure RadialTopLeft(const Colors: TGradientColors; Pattern: TBitmap);
var
  X, Y: Integer;
  Row: PRGBTripleArray;
  PreCalcY: Integer;
  PreCalcXs: array[0..180] of Integer;
begin
  Pattern.Width := 181;
  Pattern.Height := 181;
  for X := 0 to 180 do
    PreCalcXs[X] := X * X;
  for Y := 0 to 180 do
  begin
    PreCalcY := Y * Y;
    Row := PRGBTripleArray(Pattern.ScanLine[Y]);
    for X := 0 to 180 do
      Row[X] := Colors[Round(Sqrt(PreCalcXs[X] + PreCalcY))];
  end;
  { Not optimized code
  for Y := 0 to 180 do
  begin
    Row := PRGBTripleArray(Pattern.ScanLine[Y]);
    for X := 0 to 180 do
      Row[X] := Colors[Round(Sqrt(Sqr(X) + Sqr(Y)))];
  end;
  }
end;

procedure RadialTopRight(const Colors: TGradientColors; Pattern: TBitmap);
var
  X, Y: Integer;
  Row: PRGBTripleArray;
  PreCalcX, PreCalcY: Integer;
  PreCalcXs: array[0..180] of Integer;
begin
  Pattern.Width := 181;
  Pattern.Height := 181;
  for X := 0 to 180 do
  begin
    PreCalcX := 180 - X;
    PreCalcXs[X] := PreCalcX * PreCalcX;
  end;
  for Y := 0 to 180 do
  begin
    PreCalcY := Y * Y;
    Row := PRGBTripleArray(Pattern.ScanLine[Y]);
    for X := 0 to 180 do
      Row[X] := Colors[Round(Sqrt(PreCalcXs[X] + PreCalcY))];
  end;
  { Not optimized code
  for Y := 0 to 180 do
  begin
    Row := PRGBTripleArray(Pattern.ScanLine[Y]);
    for X := 0 to 180 do
      Row[X] := Colors[Round(Sqrt(Sqr(180 - X) + Sqr(Y)))];
  end;
  }
end;

procedure RadialBottomLeft(const Colors: TGradientColors; Pattern: TBitmap);
var
  X, Y: Integer;
  Row: PRGBTripleArray;
  PreCalcY: Integer;
  PreCalcXs: array[0..180] of Integer;
begin
  Pattern.Width := 181;
  Pattern.Height := 181;
  for X := 0 to 180 do
    PreCalcXs[X] := X * X;
  for Y := 0 to 180 do
  begin
    PreCalcY := 180 - Y;
    PreCalcY := PreCalcY * PreCalcY;
    Row := PRGBTripleArray(Pattern.ScanLine[Y]);
    for X := 0 to 180 do
      Row[X] := Colors[Round(Sqrt(PreCalcXs[X] + PreCalcY))];
  end;
  { Not optimized code
  for Y := 0 to 180 do
  begin
    Row := PRGBTripleArray(Pattern.ScanLine[Y]);
    for X := 0 to 180 do
      Row[X] := Colors[Round(Sqrt(Sqr(X) + Sqr(180 - Y)))];
  end;
  }
end;

procedure RadialBottomRight(const Colors: TGradientColors; Pattern: TBitmap);
var
  X, Y: Integer;
  Row: PRGBTripleArray;
  PreCalcX, PreCalcY: Integer;
  PreCalcXs: array[0..180] of Integer;
begin
  Pattern.Width := 181;
  Pattern.Height := 181;
  for X := 0 to 180 do
  begin
    PreCalcX := 180 - X;
    PreCalcXs[X] := PreCalcX * PreCalcX;
  end;
  for Y := 0 to 180 do
  begin
    PreCalcY := 180 - Y;
    PreCalcY := PreCalcY * PreCalcY;
    Row := PRGBTripleArray(Pattern.ScanLine[Y]);
    for X := 0 to 180 do
      Row[X] := Colors[Round(Sqrt(PreCalcXs[X] + PreCalcY))];
  end;
  { Not optimized code
  for Y := 0 to 180 do
  begin
    Row := PRGBTripleArray(Pattern.ScanLine[Y]);
    for X := 0 to 180 do
      Row[X] := Colors[Round(Sqrt(Sqr(180 - X) + Sqr(180 - Y)))];
  end;
  }
end;

procedure LinearHorizontal(const Colors: TGradientColors; Pattern: TBitmap);
var
  X: Integer;
  Row: PRGBTripleArray;
begin
  Pattern.Width := 256;
  Pattern.Height := 1;
  Row := PRGBTripleArray(Pattern.ScanLine[0]);
  for X := 0 to 255 do
    Row[X] := Colors[X];
end;

procedure LinearVertical(const Colors: TGradientColors; Pattern: TBitmap);
var
  Y: Integer;
  Row: PRGBTripleArray;
begin
  Pattern.Width := 1;
  Pattern.Height := 256;
  for Y := 0 to 255 do
  begin
    Row := PRGBTripleArray(Pattern.ScanLine[Y]);
    Row[0] := Colors[Y];
  end;
end;

procedure ReflectedHorizontal(const Colors: TGradientColors; Pattern: TBitmap);
var
  Y: Integer;
  Row: PRGBTripleArray;
begin
  Pattern.Width := 1;
  Pattern.Height := 512;
  for Y := 0 to 255 do
  begin
    Row := PRGBTripleArray(Pattern.ScanLine[Y]);
    Row[0] := Colors[255 - Y];
    Row := PRGBTripleArray(Pattern.ScanLine[511 - Y]);
    Row[0] := Colors[255 - Y];
  end;
end;

procedure ReflectedVertical(const Colors: TGradientColors; Pattern: TBitmap);
var
  X: Integer;
  Row: PRGBTripleArray;
begin
  Pattern.Width := 512;
  Pattern.Height := 1;
  Row := PRGBTripleArray(Pattern.ScanLine[0]);
  for X := 0 to 255 do
  begin
    Row[X] := Colors[255 - X];
    Row[511 - X] := Colors[255 - X];
  end;
end;

procedure DiagonalLinearForward(const Colors: TGradientColors; Pattern: TBitmap);
var
  X, Y: Integer;
  Row: PRGBTripleArray;
begin
  Pattern.Width := 128;
  Pattern.Height := 129;
  for Y := 0 to 128 do
  begin
    Row := PRGBTripleArray(Pattern.ScanLine[Y]);
    for X := 0 to 127 do
      Row[X] := Colors[X + Y];
  end;
end;

procedure DiagonalLinearBackward(const Colors: TGradientColors; Pattern: TBitmap);
var
  X, Y: Integer;
  Row: PRGBTripleArray;
begin
  Pattern.Width := 128;
  Pattern.Height := 129;
  for Y := 0 to 128 do
  begin
    Row := PRGBTripleArray(Pattern.ScanLine[Y]);
    for X := 0 to 127 do
      Row[X] := Colors[127 + (Y - X)];
  end;
end;

procedure DiagonalReflectedForward(const Colors: TGradientColors; Pattern: TBitmap);
var
  X, Y: Integer;
  Row: PRGBTripleArray;
begin
  Pattern.Width := 256;
  Pattern.Height := 256;
  for Y := 0 to 255 do
  begin
    Row := PRGBTripleArray(Pattern.ScanLine[Y]);
    for X := 0 to 255 do
      if X + Y < 255 then
        Row[X] := Colors[255 - (X + Y)]
      else
        Row[X] := Colors[(Y + X) - 255];
  end;
end;

procedure DiagonalReflectedBackward(const Colors: TGradientColors; Pattern: TBitmap);
var
  X, Y: Integer;
  Row: PRGBTripleArray;
begin
  Pattern.Width := 256;
  Pattern.Height := 256;
  for Y := 0 to 255 do
  begin
    Row := PRGBTripleArray(Pattern.ScanLine[Y]);
    for X := 0 to 255 do
      if X > Y then
        Row[X] := Colors[X - Y]
      else
        Row[X] := Colors[Y - X];
  end;
end;

procedure ArrowLeft(const Colors: TGradientColors; Pattern: TBitmap);
var
  X, Y: Integer;
  Row: PRGBTripleArray;
begin
  Pattern.Width := 129;
  Pattern.Height := 256;
  for Y := 0 to 127 do
  begin
    Row := PRGBTripleArray(Pattern.ScanLine[Y]);
    for X := 0 to 128 do
      Row[X] := Colors[255 - (X + Y)];
  end;
  for Y := 128 to 255 do
  begin
    Row := PRGBTripleArray(Pattern.ScanLine[Y]);
    for X := 0 to 128 do
      Row[X] := Colors[Y - X];
  end;
end;

procedure ArrowRight(const Colors: TGradientColors; Pattern: TBitmap);
var
  X, Y: Integer;
  Row: PRGBTripleArray;
begin
  Pattern.Width := 129;
  Pattern.Height := 256;
  for Y := 0 to 127 do
  begin
    Row := PRGBTripleArray(Pattern.ScanLine[Y]);
    for X := 0 to 128 do
      Row[X] := Colors[(X - Y) + 127];
  end;
  for Y := 128 to 255 do
  begin
    Row := PRGBTripleArray(Pattern.ScanLine[Y]);
    for X := 0 to 128 do
      Row[X] := Colors[(X + Y) - 128];
  end;
end;

procedure ArrowUp(const Colors: TGradientColors; Pattern: TBitmap);
var
  X, Y: Integer;
  Row: PRGBTripleArray;
begin
  Pattern.Width := 256;
  Pattern.Height := 129;
  for Y := 0 to 128 do
  begin
    Row := PRGBTripleArray(Pattern.ScanLine[Y]);
    for X := 0 to 127 do
      Row[X] := Colors[255 - (X + Y)];
    for X := 128 to 255 do
      Row[X] := Colors[X - Y];
  end;
end;

procedure ArrowDown(const Colors: TGradientColors; Pattern: TBitmap);
var
  X, Y: Integer;
  Row: PRGBTripleArray;
begin
  Pattern.Width := 256;
  Pattern.Height := 129;
  for Y := 0 to 128 do
  begin
    Row := PRGBTripleArray(Pattern.ScanLine[Y]);
    for X := 0 to 127 do
      Row[X] := Colors[127 + (Y - X)];
    for X := 128 to 255 do
      Row[X] := Colors[(X + Y) - 128];
  end;
end;

procedure Diamond(const Colors: TGradientColors; Pattern: TBitmap);
var
  X, Y: Integer;
  Row: PRGBTripleArray;
begin
  Pattern.Width := 256;
  Pattern.Height := 256;
  for Y := 0 to 127 do
  begin
    Row := PRGBTripleArray(Pattern.ScanLine[Y]);
    for X := 0 to 127 do
      Row[X] := Colors[255 - (X + Y)];
    for X := 128 to 255 do
      Row[X] := Colors[X - Y];
  end;
  for Y := 128 to 255 do
  begin
    Row := PRGBTripleArray(Pattern.ScanLine[Y]);
    for X := 0 to 127 do
      Row[X] := Colors[Y - X];
    for X := 128 to 255 do
      Row[X] := Colors[(X + Y) - 255];
  end;
end;

procedure Butterfly(const Colors: TGradientColors; Pattern: TBitmap);
var
  X, Y: Integer;
  Row: PRGBTripleArray;
begin
  Pattern.Width := 256;
  Pattern.Height := 256;
  for Y := 0 to 127 do
  begin
    Row := PRGBTripleArray(Pattern.ScanLine[Y]);
    for X := 0 to 127 do
      Row[X] := Colors[(X - Y) + 128];
    for X := 128 to 255 do
      Row[X] := Colors[383 - (X + Y)];
  end;
  for Y := 128 to 255 do
  begin
    Row := PRGBTripleArray(Pattern.ScanLine[Y]);
    for X := 0 to 127 do
      Row[X] := Colors[(X + Y) - 128];
    for X := 128 to 255 do
      Row[X] := Colors[128 + (Y - X)];
  end;
end;

{ TGradient }

type
  TPatternBuilder = procedure(const Colors: TGradientColors; Pattern: TBitmap);

const
  PatternBuilder: array[TGradientStyle] of TPatternBuilder = (nil,
    RadialCentral, RadialTop, RadialBottom, RadialLeft, RadialRight,
    RadialTopLeft, RadialTopRight, RadialBottomLeft, RadialBottomRight,
    LinearHorizontal, LinearVertical, ReflectedHorizontal, ReflectedVertical,
    DiagonalLinearForward, DiagonalLinearBackward, DiagonalReflectedForward,
    DiagonalReflectedBackward, ArrowLeft, ArrowRight, ArrowUp, ArrowDown,
    Diamond, Butterfly);

constructor TGradient.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csOpaque];
  Width := 100;
  Height := 100;
  fColorBegin := clWhite;
  fColorEnd := clBtnFace;
  fStyle := gsRadialC;
  fShift := 0;
  fRotation := 0;
  fReverse := False;
  fUseSysColors := False;
  fPattern := TBitmap.Create;
  fPattern.PixelFormat := pf24bit;
  UpdatePattern;
end;

destructor TGradient.Destroy;
begin
  fPattern.Free;
  inherited Destroy;
end;

procedure TGradient.Loaded;
begin
  inherited Loaded;
  UpdatePattern;
end;

procedure TGradient.Paint;
begin
  if not Dirty then
    Canvas.StretchDraw(ClientRect, Pattern);
end;

procedure TGradient.BeginUpdate;
begin
  Inc(UpdateCount);
end;

procedure TGradient.EndUpdate;
begin
  Dec(UpdateCount);
  if (UpdateCount = 0) and UpdatePended then
    UpdatePattern;
end;

function TGradient.CopyPatternTo(Bitmap: TBitmap): Boolean;
begin
  Result := False;
  if not Dirty and (UpdateCount = 0) and Assigned(Bitmap) then
  begin
    Bitmap.Assign(Pattern);
    Result := True;
  end;
end;

procedure TGradient.InvalidatePattern;
begin
  UpdatePattern;
end;

procedure TGradient.WMSETTINGCHANGE(var Message: TMessage);
begin
  inherited;
  if UseSysColors then
    UpdateSysColors;
end;

procedure TGradient.SetColorBegin(Value: TColor);
begin
  if fColorBegin <> Value then
  begin
    fColorBegin := Value;
    fUseSysColors := False;
    UpdatePattern;
  end;
end;

procedure TGradient.SetColorEnd(Value: TColor);
begin
  if fColorEnd <> Value then
  begin
    fColorEnd := Value;
    fUseSysColors := False;
    UpdatePattern;
  end;
end;

procedure TGradient.SetUseSysColors(Value: Boolean);
begin
  if fUseSysColors <> Value then
  begin
    fUseSysColors := Value;
    if fUseSysColors then
      UpdateSysColors;
  end;
end;

procedure TGradient.SetStyle(Value: TGradientStyle);
begin
  if fStyle <> Value then
  begin
    fStyle := Value;
    UpdatePattern;
  end;
end;

procedure TGradient.SetShift(Value: TGradientShift);
begin
  if Value < Low(TGradientShift) then
    Value := Low(TGradientShift)
  else if Value > High(TGradientShift) then
    Value := High(TGradientShift);

  if fShift <> Value then
  begin
    fShift := Value;
    UpdatePattern;
  end;
end;

procedure TGradient.SetRotation(Value: TGradientRotation);
begin
  if Value < Low(TGradientRotation) then
    Value := Low(TGradientRotation)
  else if Value > High(TGradientRotation) then
    Value := High(TGradientRotation);

  if fRotation <> Value then
  begin
    fRotation := Value;
    UpdatePattern;
  end;
end;

procedure TGradient.SetReverse(Value: Boolean);
begin
  if fReverse <> Value then
  begin
    fReverse := Value;
    UpdatePattern;
  end;
end;

function TGradient.IsColorBeginSaved: Boolean;
begin
  Result := not UseSysColors and (ColorBegin <> clWhite);
end;

function TGradient.IsColorEndSaved: Boolean;
begin
  Result := not UseSysColors and (ColorBegin <> clBtnFace);
end;

procedure TGradient.UpdateSysColors;
{$IFDEF DELPHI3}
const
  COLOR_GRADIENTACTIVECAPTION = 27;
{$ENDIF}
begin
  BeginUpdate;
  try
    ColorBegin := GetSysColor(COLOR_ACTIVECAPTION);
    try
      ColorEnd := GetSysColor(COLOR_GRADIENTACTIVECAPTION);
      fUseSysColors := True;
    except
      // This windows version doesn't support gradient colors...
      ColorEnd := ColorBegin;
      fUseSysColors := False;
    end;
  finally
    EndUpdate;
  end;
end;

procedure TGradient.UpdatePattern;
var
  Colors: TGradientColors;
  dRed, dGreen, dBlue: Integer;
  RGBColor1, RGBColor2: TColor;
  RGB1, RGB2: TRGBTriple;
  UpdatedRect: TRect;
  Index: Integer;
  M: Integer;
begin
  UpdatePended := True;

  if (csLoading in ComponentState) or (UpdateCount <> 0) then Exit;

  UpdatePended := False;

  if Reverse then
  begin
    RGBColor1 := ColorToRGB(ColorEnd);
    RGBColor2 := ColorToRGB(ColorBegin);
  end
  else
  begin
    RGBColor1 := ColorToRGB(ColorBegin);
    RGBColor2 := ColorToRGB(ColorEnd);
  end;

  RGB1.rgbtRed := GetRValue(RGBColor1);
  RGB1.rgbtGreen := GetGValue(RGBColor1);
  RGB1.rgbtBlue := GetBValue(RGBColor1);

  RGB2.rgbtRed := GetRValue(RGBColor2);
  RGB2.rgbtGreen := GetGValue(RGBColor2);
  RGB2.rgbtBlue := GetBValue(RGBColor2);

  if Shift > 0 then
  begin
    Inc(RGB1.rgbtRed, MulDiv(RGB2.rgbtRed - RGB1.rgbtRed, Shift, 100));
    Inc(RGB1.rgbtGreen, MulDiv(RGB2.rgbtGreen - RGB1.rgbtGreen, Shift, 100));
    Inc(RGB1.rgbtBlue, MulDiv(RGB2.rgbtBlue - RGB1.rgbtBlue, Shift, 100));
  end
  else if Shift < 0 then
  begin
    Inc(RGB2.rgbtRed, MulDiv(RGB2.rgbtRed - RGB1.rgbtRed, Shift, 100));
    Inc(RGB2.rgbtGreen, MulDiv(RGB2.rgbtGreen - RGB1.rgbtGreen, Shift, 100));
    Inc(RGB2.rgbtBlue, MulDiv(RGB2.rgbtBlue - RGB1.rgbtBlue, Shift, 100));
  end;

  dRed := RGB2.rgbtRed - RGB1.rgbtRed;
  dGreen := RGB2.rgbtGreen - RGB1.rgbtGreen;
  dBlue := RGB2.rgbtBlue - RGB1.rgbtBlue;

  M := MulDiv(255, Rotation, 100);
  if M = 0 then
    for Index := 0 to 255 do
      with Colors[Index] do
      begin
        rgbtRed := RGB1.rgbtRed + (Index * dRed) div 255;
        rgbtGreen := RGB1.rgbtGreen + (Index * dGreen) div 255;
        rgbtBlue := RGB1.rgbtBlue + (Index * dBlue) div 255;
      end
  else if M > 0 then
  begin
    M := 255 - M;
    for Index := 0 to M - 1 do
      with Colors[Index] do
      begin
        rgbtRed := RGB1.rgbtRed + (Index * dRed) div M;
        rgbtGreen := RGB1.rgbtGreen + (Index * dGreen) div M;
        rgbtBlue := RGB1.rgbtBlue + (Index * dBlue) div M;
      end;
    for Index := M to 255 do
      with Colors[Index] do
      begin
        rgbtRed := RGB1.rgbtRed + ((255 - Index) * dRed) div (255 - M);
        rgbtGreen := RGB1.rgbtGreen + ((255 - Index) * dGreen) div (255 - M);
        rgbtBlue := RGB1.rgbtBlue + ((255 - Index) * dBlue) div (255 - M);
      end;
  end
  else if M < 0 then
  begin
    M := -M;
    for Index := 0 to M do
      with Colors[Index] do
      begin
        rgbtRed := RGB2.rgbtRed - (Index * dRed) div M;
        rgbtGreen := RGB2.rgbtGreen - (Index * dGreen) div M;
        rgbtBlue := RGB2.rgbtBlue - (Index * dBlue) div M;
      end;
    for Index := M + 1 to 255 do
      with Colors[Index] do
      begin
        rgbtRed := RGB2.rgbtRed - ((255 - Index) * dRed) div (255 - M);
        rgbtGreen := RGB2.rgbtGreen - ((255 - Index) * dGreen) div (255 - M);
        rgbtBlue := RGB2.rgbtBlue - ((255 - Index) * dBlue) div (255 - M);
      end;
  end;

  Dirty := True;
  try
    if @PatternBuilder[Style] <> nil then
      PatternBuilder[Style](Colors, Pattern)
    else if Assigned(fOnCustom) then
      fOnCustom(Self, Colors, Pattern)
    else
    begin
      Pattern.Width := 2;
      Pattern.Height := 2;
      Pattern.Canvas.Pixels[0, 0] := RGBColor1;
      Pattern.Canvas.Pixels[0, 1] := RGBColor2;
      Pattern.Canvas.Pixels[1, 0] := RGBColor2;
      Pattern.Canvas.Pixels[1, 1] := RGBColor1;
    end;
  finally
    Dirty := False;
  end;

  if (Parent <> nil) and Parent.HandleAllocated then
  begin
    UpdatedRect := BoundsRect;
    InvalidateRect(Parent.Handle, @UpdatedRect, False);
    if csDesigning in ComponentState then Parent.Update;
  end
  else
    Invalidate;
end;

procedure Register;
begin
  RegisterComponents('Delphi Area', [TGradient]);
end;

end.