{------------------------------------------------------------------------------}
{                                                                              }
{  TSysImageList v1.31                                                         }
{  by Kambiz R. Khojasteh                                                      }
{                                                                              }
{  kambiz@delphiarea.com                                                       }
{  http://www.delphiarea.com                                                   }
{                                                                              }
{------------------------------------------------------------------------------}

{$I DELPHIAREA.INC}

unit SysImg;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  {$IFDEF DELPHI4_UP} ImgList {$ELSE} ComCtrls {$ENDIF};

type

  TIconSize = (isSmallIcons, isLargeIcons);
  TIconSet = (isSystem, isToolbar);

  {$IFDEF DELPHI4_UP}
  TSysImageList = class(TCustomImageList)
  {$ELSE}
  TSysImageList = class(TImageList)
  {$ENDIF}
  private
    fIconSet: TIconSet;
    fIconSize: TIconSize;
    TBHandle: THandle;
    procedure SetIconSet(Value: TIconSet);
    procedure SetIconSize(Value: TIconSize);
  protected
    procedure Loaded; override;
    procedure DefineProperties(Filer: TFiler); override;
    procedure UpdateHandle;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function ImageIndexOf(const Path: String; OpenIcon: Boolean
      {$IFDEF DELPHI4_UP} = False {$ENDIF}): Integer;
  published
    Property BkColor default clNone;
    Property BlendColor default clNone;
    property DrawingStyle default dsTransparent;
    property IconSet: TIconSet read fIconSet write SetIconSet default isSystem;
    property IconSize: TIconSize read fIconSize write SetIconSize default isSmallIcons;
    property OnChange;
  end;

procedure Register;

implementation

uses
  ShellAPI, ShlObj, ActiveX, CommCtrl {$IFDEF DELPHI4_UP}, ComCtrls {$ENDIF};

procedure Register;
begin
  RegisterComponents('Delphi Area', [TSysImageList]);
end;

{ TCustomSysImageList }

const
  IconSizeFlags: array[TIconSize] of Word = (SHGFI_SMALLICON, SHGFI_LARGEICON);
  OpenIconFlags: array[Boolean] of Word = (0, SHGFI_OPENICON);
  IconSizeStdIDs: array[TIconSize] of Integer = (IDB_STD_SMALL_COLOR, IDB_STD_LARGE_COLOR);
  IconSizeViewIDs: array[TIconSize] of Integer = (IDB_VIEW_SMALL_COLOR, IDB_VIEW_LARGE_COLOR);
  IconSizeHistIDs: array[TIconSize] of Integer = (IDB_HIST_SMALL_COLOR, IDB_HIST_LARGE_COLOR);

constructor TSysImageList.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  fIconSet := isSystem;
  fIconSize := isSmallIcons;
  ShareImages := True;
  DrawingStyle := dsTransparent;
  BlendColor := clNone;
  BkColor := clNone;
  UpdateHandle;
end;

destructor TSysImageList.Destroy;
begin
  if TBHandle <> 0 then
    DestroyWindow(TBHandle);
  inherited Destroy;
end;

procedure TSysImageList.Loaded;
begin
  inherited Loaded;
  UpdateHandle;
end;

procedure TSysImageList.DefineProperties(Filer: TFiler);
begin
  Handle := 0;
  inherited DefineProperties(Filer);
  UpdateHandle;
end;

procedure TSysImageList.SetIconSet(Value: TIconSet);
begin
  if fIconSet <> Value then
  begin
    fIconSet := Value;
    UpdateHandle;
  end;
end;

procedure TSysImageList.SetIconSize(Value: TIconSize);
begin
  if fIconSize <> Value then
  begin
    fIconSize := Value;
    UpdateHandle;
  end;
end;

procedure TSysImageList.UpdateHandle;
var
  FileInfo: TShFileInfo;
  TBAddBitmap: TTBAddBitmap;
begin
  if not (csLoading in ComponentState) then
  begin
    if TBHandle <> 0 then
    begin
      DestroyWindow(TBHandle);
      TBHandle := 0;
    end;
    case IconSet of
      isSystem:
      begin
        Handle := ShGetFileInfo(PChar(Application.ExeName), 0, FileInfo,
          SizeOf(FileInfo), IconSizeFlags[IconSize] or SHGFI_SYSICONINDEX or SHGFI_ICONLOCATION)
      end;
      isToolbar:
      begin
        TBHandle := CreateWindow(TOOLBARCLASSNAME, nil, 0, 0, 0, 0, 0, 0, 0, HInstance, nil);
        TBAddBitmap.hInst := HINST_COMMCTRL;
        TBAddBitmap.nID := IconSizeStdIDs[IconSize];
        SendMessage(TBHandle, TB_ADDBITMAP, 1000, Integer(@TBAddBitmap));
        TBAddBitmap.nID := IconSizeViewIDs[IconSize];
        SendMessage(TBHandle, TB_ADDBITMAP, 1000, Integer(@TBAddBitmap));
        TBAddBitmap.nID := IconSizeHistIDs[IconSize];
        SendMessage(TBHandle, TB_ADDBITMAP, 1000, Integer(@TBAddBitmap));
        Handle := SendMessage(TBHandle, TB_GETIMAGELIST, 0, 0);
      end;
    end;
    BkColor := clNone;
  end;
end;

function TSysImageList.ImageIndexOf(const Path: String; OpenIcon: Boolean): Integer;

  function GetUniqueFileName: String;
  var
    TempPath: array[0..255] of Char;
    TempFile: array[0..255] of Char;
  begin
    GetTempPath(SizeOf(TempPath), TempPath);
    GetTempFileName(TempPath, 'SIL', 0, TempFile);
    Result := TempFile;
  end;

var
  FileInfo: TShFileInfo;
  DesktopFolder: IShellFolder;
  PIDL: PItemIDList;
  Malloc: IMalloc;
  NumChars, Flags: {$IFDEF DELPHI4_UP} Cardinal {$ELSE} LongInt {$ENDIF};
  WidePath: PWideChar;
  FileExt, TempFileName: String;
  TempFile: THandle;
begin
  Result := 0;
  Flags := 0;
  WidePath := StringToOleStr(Path);
  NumChars := Length(WidePath);
  SHGetDesktopFolder(DesktopFolder);
  DesktopFolder.ParseDisplayName(0, nil, WidePath, NumChars, PIDL, Flags);
  // If PIDL is obtained, we use it for finding the image index. When path is not
  // a URL or an existing file, PIDL could be null.
  if PIDL <> nil then
  begin
    FillChar(FileInfo, SizeOf(FileInfo), 0);
    ShGetFileInfo(PChar(PIDL), 0, FileInfo, SizeOf(FileInfo), SHGFI_PIDL or
      SHGFI_SYSICONINDEX or IconSizeFlags[fIconSize] or OpenIconFlags[OpenIcon]);
    if FileInfo.iIcon > 0 then Result := FileInfo.iIcon;
    ShGetMalloc(Malloc);
    Malloc.Free(PIDL);
  end
  else
  begin
    // If PIDL is null, we try to find the image index by using the file type
    FileExt := ExtractFileExt(Path);
    FillChar(FileInfo, SizeOf(FileInfo), 0);
    ShGetFileInfo(PChar('*.' + FileExt), FILE_ATTRIBUTE_NORMAL, FileInfo,
      SizeOf(FileInfo), SHGFI_USEFILEATTRIBUTES or SHGFI_SYSICONINDEX or
      IconSizeFlags[IconSize] or OpenIconFlags[OpenIcon]);
    if FileInfo.iIcon > 0 then
      Result := FileInfo.iIcon
    else if not FileExists(Path) then
    begin
      // If using file type failed and path does not point to an existing file,
      // we create a temporary file as type of the file. If there is any icon
      // associated to the type, we will get it.
      TempFileName := GetUniqueFileName + FileExt;
      TempFile := FileCreate(TempFileName);
      if TempFile > 0 then
        FileClose(TempFile);
      if FileExists(TempFileName) then
        try
          Result := ImageIndexOf(TempFileName, OpenIcon);
        finally
          DeleteFile(TempFileName);
        end;
    end;
  end;
end;

end.
